
#include "pixeltoaster/PixelToaster.h"
using namespace PixelToaster;

#include "vector"

#include "MiniScheduler/MiniScheduler.h"
#include "MiniScheduler/ParallelFor.h"

#include "entitysystem.h"

const int width = 800;
const int height = 600;

const int numberBodies = 300;

/**
 * Linear Congruential Random Number Generator (approx cycle 2^^32) for Floatingpoint Values
 */
class LCRGF {
public:
   LCRGF() { mCurrentValue = 1; }
   float randf1to1() {
      mCurrentValue *= 16807;
      return ((float)mCurrentValue) / (float)0x80000000;
   }
private:
   int mCurrentValue;
};

struct CompPosition3D : public Component {
   static const FamilyId familyId = 1;
   CompPosition3D() : x(0), y(0), z(0) {}
   float x,y,z;
};
struct CompPosition3DTemporary : public Component {
   static const FamilyId familyId = 2;
   CompPosition3DTemporary() : x(0), y(0), z(0) {}
   float x,y,z;
};
struct CompVelocity : public Component {
   static const FamilyId familyId = 3;
   CompVelocity() : vx(0), vy(0), vz(0) {}
   float vx,vy,vz;
};
struct CompPosition2D : public Component {
   static const FamilyId familyId = 4;
   CompPosition2D() : x(0), y(0) {}
   int x,y;
};

//Tasks for parallel work
class NBodyIntegrationStep : public ParallelForTask {
public:
   NBodyIntegrationStep(std::vector<Entity*> &entities, float _dT)
      : mEntities(entities)
      , dT(_dT) {}
   //ParallelForTask:
   bool doRange( WorkerThread* pThread, const ParallelRange& range ) const {
      for(int i=range.mBegin; i<range.mEnd; ++i) {
         float accX = 0.0f;
         float accY = 0.0f;
         float accZ = 0.0f;
         CompPosition3D *myBodyPos = mEntities[i]->getAs<CompPosition3D>();
         for(unsigned int j=0; j<mEntities.size(); ++j) {
            CompPosition3D *otherBodyPos = mEntities[j]->getAs<CompPosition3D>();
            float deltaX = otherBodyPos->x - myBodyPos->x;
            float deltaY = otherBodyPos->y - myBodyPos->y;
            float deltaZ = otherBodyPos->z - myBodyPos->z;
            float invR = 1.0f / sqrt(deltaX*deltaX + deltaY*deltaY + deltaZ*deltaZ + 0.000061f); //epsilon
            float invR3 = invR*invR*invR;
            float f = 10.0f * invR3;         //mass
            accX += f * deltaX;
            accY += f * deltaY;
            accZ += f * deltaZ;
         }
         CompPosition3DTemporary *myBodyPosTemp = mEntities[i]->getAs<CompPosition3DTemporary>();
         CompVelocity *myBodyVelocity = mEntities[i]->getAs<CompVelocity>();
         myBodyPosTemp->x = myBodyPos->x + dT*myBodyVelocity->vx + 0.5f*dT*dT*accX;
         myBodyPosTemp->y = myBodyPos->y + dT*myBodyVelocity->vy + 0.5f*dT*dT*accY;
         myBodyPosTemp->z = myBodyPos->z + dT*myBodyVelocity->vz + 0.5f*dT*dT*accZ;
         myBodyVelocity->vx += dT*accX;
         myBodyVelocity->vy += dT*accY;
         myBodyVelocity->vz += dT*accZ;
      }
      return true;
   }
protected:
   float dT;
   std::vector<Entity*> &mEntities;
};
class NBodyCopyTempPos : public ParallelForTask {
public:
   NBodyCopyTempPos(std::vector<Entity*> &entities)
      : mEntities(entities) {}
   //ParallelForTask:
   bool doRange( WorkerThread* pThread, const ParallelRange& range ) const {
      for(int i=range.mBegin; i<range.mEnd; ++i) {
         CompPosition3D *bodyPos = mEntities[i]->getAs<CompPosition3D>();
         CompPosition3DTemporary *bodyPosTemp = mEntities[i]->getAs<CompPosition3DTemporary>();
         bodyPos->x = bodyPosTemp->x;
         bodyPos->y = bodyPosTemp->y;
         bodyPos->z = bodyPosTemp->z;
      }
      return true;
   }
protected:
   std::vector<Entity*> &mEntities;
};
class Project3Dto2D : public ParallelForTask {
public:
   Project3Dto2D(std::vector<Entity*> &entities)
      : mEntities(entities) {}
   //ParallelForTask:
   bool doRange( WorkerThread* pThread, const ParallelRange& range ) const {
      for(int i=range.mBegin; i<range.mEnd; ++i) {
         CompPosition3D *bodyPos3D = mEntities[i]->getAs<CompPosition3D>();
         CompPosition2D *bodyPos2D = mEntities[i]->getAs<CompPosition2D>();

         bodyPos2D->x = (int)((width/2) + bodyPos3D->x * 200 / (bodyPos3D->z + 1.0f));
         bodyPos2D->y = (int)((height/2) + bodyPos3D->y * 200 / (bodyPos3D->z + 1.0f));
      }
      return true;
   }
protected:
   std::vector<Entity*> &mEntities;
};

//=================================================================
void doBenchmark() {
   const int numberEntities = 2000;
   Timer timer;

   EntitySystem entitySystem;
   Entity entities[numberEntities];
   unsigned int i;
   for(i=0; i<numberEntities; ++i) {
      CompPosition3D *compPos3D = new CompPosition3D();
      CompPosition3DTemporary *compPos3DTemp = new CompPosition3DTemporary();
      CompVelocity *compVelocity = new CompVelocity();
      CompPosition2D *compPos2D = new CompPosition2D();
      entitySystem.addComponent(&entities[i], compPos3D);
      entitySystem.addComponent(&entities[i], compPos3DTemp);
      entitySystem.addComponent(&entities[i], compVelocity);
      entitySystem.addComponent(&entities[i], compPos2D);
   }

   i=0;
   timer.reset();
   while(i<10000) {
      for(unsigned int j=0; j<numberEntities; ++j) {
         CompPosition3D *temp1 = entities[j].getAs<CompPosition3D>();
         temp1->x = temp1->y = temp1->z = i;
         CompPosition3DTemporary *temp2 = entities[j].getAs<CompPosition3DTemporary>();
         temp2->x = temp2->y = temp2->z = i;
         CompPosition2D *temp3 = entities[j].getAs<CompPosition2D>();
         temp3->x = temp3->y = i;
         CompVelocity *temp4 = entities[j].getAs<CompVelocity>();
         temp4->vx = temp4->vy = temp4->vz = i;
      }
      ++i;
   }
   double delta = timer.delta();
   printf("benchmark took %gs\n", delta);
}

int main() {
   Display display( "Parallel NBody with Entity System", width, height, Output::Default, Mode::TrueColor );
   std::vector<TrueColorPixel> pixels( width * height );

   /* dx9 does this aswell */
   _controlfp(_PC_24, MCW_PC);   //single precision
   _controlfp(_RC_NEAR, MCW_RC); //round-to-nearest

   //doBenchmark();

   TaskPool tasks;
   tasks.start(0);

   Timer timer;

   LCRGF lcrgf;

   EntitySystem entitySystem;
   Entity entities[numberBodies];
   for(unsigned int i=0; i<numberBodies; ++i) {
      CompPosition3D *compPos3D = new CompPosition3D();
      compPos3D->x = lcrgf.randf1to1();
      compPos3D->y = lcrgf.randf1to1();
      compPos3D->z = lcrgf.randf1to1();
      CompPosition3DTemporary *compPos3DTemp = new CompPosition3DTemporary();
      CompVelocity *compVelocity = new CompVelocity();
      CompPosition2D *compPos2D = new CompPosition2D();
      entitySystem.addComponent(&entities[i], compPos3D);
      entitySystem.addComponent(&entities[i], compPos3DTemp);
      entitySystem.addComponent(&entities[i], compVelocity);
      entitySystem.addComponent(&entities[i], compPos2D);
   }

   double delta = 0.01;
   while ( display.open() ) {
      timer.reset();

      std::vector<Entity*> entitiesWithPosition;
      entitySystem.getEntities<CompPosition3D>(entitiesWithPosition);

//      for(int i=0; i<width*height; ++i) {          //clear screen
//         pixels[i].r = pixels[i].g = pixels[i].b = 0;
//      }
      //its faster to delete the pixels on the old position...
      for(unsigned int i=0; i<entitiesWithPosition.size(); ++i) {
         CompPosition2D *bodyPos = entitiesWithPosition[i]->getAs<CompPosition2D>();
         if(bodyPos->x >= 0 && bodyPos->y >= 0 && bodyPos->x < width && bodyPos->y < height) {
            int index = bodyPos->x + bodyPos->y * width;
            pixels[index].r = pixels[index].g = pixels[index].b = 0;
         }
      }

      WorkerThread::profilerStartFrame();
      bool bOk;

      NBodyIntegrationStep integrationStep(entitiesWithPosition, (float)0.001);
      bOk = ParallelFor(&integrationStep, ParallelRange(0, entitiesWithPosition.size(), 20));
      assert(bOk);

      NBodyCopyTempPos copyTempPos(entitiesWithPosition);
      bOk = ParallelFor(&copyTempPos, ParallelRange(0, entitiesWithPosition.size(), 20));
      assert(bOk);

      Project3Dto2D project3Dto2D(entitiesWithPosition);
      bOk = ParallelFor(&project3Dto2D, ParallelRange(0, entitiesWithPosition.size(), 20));
      assert(bOk);

      WorkerThread::profilerEndFrame();
      WorkerThread::drawProfiler(&tasks, (unsigned int*)pixels.data(), width, height);

      //Draw
      for(unsigned int i=0; i<entitiesWithPosition.size(); ++i) {
         CompPosition2D *bodyPos = entitiesWithPosition[i]->getAs<CompPosition2D>();
         if(bodyPos->x >= 0 && bodyPos->y >= 0 && bodyPos->x < width && bodyPos->y < height) {
            int index = bodyPos->x + bodyPos->y * width;
            pixels[index].r = pixels[index].g = pixels[index].b = 200;
         }
      }

      delta = timer.delta();
      //printf( "one frame took %gs\n", delta );

      display.update( pixels.data() );
   }
   tasks.stop();
}
